import logging
import os

supported_providers = ["github", "travis", "appveyor"]


def get_continuous_integration_parameters():
    """
    Return a dictionary with information on a continuous integration build
    inferred from environment variables.

    The following dictionary keys are set (all values are also strings):
    - provider: name of CI provider, such as "travis" or "appveyor".
    - repo_slug: owner/name of the source code repository, i.e. "manubot/rootstock"
    - repo_owner: owner from repo_slug, i.e. "manubot"
    - repo_name: name from repo_slug, i.e. "rootstock"
    - commit: git commit being evaluated by the CI build
        For pull requests, CI services often build a merge commit
        with the default branch rather than the commit that was
        added to the pull request. In these cases, commit refers
        to this merge commit.
    - triggering_commit: git commit that triggered the CI build.
    - build_url: URL for the webpage with build details
    - job_url: URL for the webpage with job details

    GitHub Actions does not set an environment variable with triggering_commit
    for pull requests. Therefore, set the following environment variable in your workflow:

    ```yaml
    env:
      GITHUB_PULL_REQUEST_SHA: ${{ github.event.pull_request.head.sha }}
    ```
    """
    if os.getenv("GITHUB_ACTIONS", "false") == "true":
        # https://git.io/JvUf7
        repo_slug = os.environ["GITHUB_REPOSITORY"]
        repo_owner, repo_name = repo_slug.split("/")
        run_id = os.environ["GITHUB_RUN_ID"]
        # GITHUB_SHA for pull_request event: Last merge commit on the GITHUB_REF branch
        # GITHUB_SHA for push event: Commit pushed, unless deleting a branch (when it's the default branch)
        # https://git.io/JvUfd
        github_sha = os.environ["GITHUB_SHA"]
        ci_params = {
            "provider": "github",
            "repo_slug": repo_slug,
            "repo_owner": repo_owner,
            "repo_name": repo_name,
            "commit": github_sha,
            "triggering_commit": os.getenv("GITHUB_PULL_REQUEST_SHA") or github_sha,
            "build_url": f"https://github.com/{repo_slug}/commit/{github_sha}/checks",
            "job_url": f"https://github.com/{repo_slug}/actions/runs/{run_id}",
        }
        return ci_params

    if os.getenv("TRAVIS", "false") == "true":
        # https://docs.travis-ci.com/user/environment-variables/
        repo_slug = os.environ["TRAVIS_REPO_SLUG"]
        repo_owner, repo_name = repo_slug.split("/")
        return {
            "provider": "travis",
            "repo_slug": repo_slug,
            "repo_owner": repo_owner,
            "repo_name": repo_name,
            "commit": os.environ["TRAVIS_COMMIT"],
            "triggering_commit": os.getenv("TRAVIS_PULL_REQUEST_SHA")
            or os.environ["TRAVIS_COMMIT"],
            "build_url": os.environ["TRAVIS_BUILD_WEB_URL"],
            "job_url": os.environ["TRAVIS_JOB_WEB_URL"],
        }

    if os.getenv("APPVEYOR", "false").lower() == "true":
        # https://www.appveyor.com/docs/environment-variables/
        repo_slug = os.environ["APPVEYOR_REPO_NAME"]
        repo_owner, repo_name = repo_slug.split("/")
        provider_url = "{APPVEYOR_URL}/project/{APPVEYOR_ACCOUNT_NAME}/{APPVEYOR_PROJECT_SLUG}".format(
            **os.environ
        )
        build_url = f"{provider_url}/builds/{os.environ['APPVEYOR_BUILD_ID']}"
        return {
            "provider": "appveyor",
            "provider_account": os.environ["APPVEYOR_ACCOUNT_NAME"],
            "repo_slug": repo_slug,
            "repo_owner": repo_owner,
            "repo_name": repo_name,
            "commit": os.environ["APPVEYOR_REPO_COMMIT"],
            "triggering_commit": os.getenv("APPVEYOR_PULL_REQUEST_HEAD_COMMIT")
            or os.environ["APPVEYOR_REPO_COMMIT"],
            "build_url": build_url,
            "job_url": f"{provider_url}/build/job/{os.environ['APPVEYOR_JOB_ID']}",
            "artifact_url": f"{build_url}/artifacts",
        }

    if os.getenv("CI", "false").lower() == "true":
        logging.warning(
            "Detected CI environment variable, but get_continuous_integration_parameters "
            "did not detect environment variables for a supported CI provider. "
            "Supported providers are: {}".format(", ".join(supported_providers))
        )
    return None



import json
import logging
import urllib.parse
import urllib.request
from typing import Any, Callable, Optional

import requests

from manubot.util import get_manubot_user_agent

from .handlers import Handler
from .pubmed import get_pubmed_ids_for_doi


class Handler_DOI(Handler):

    standard_prefix = "doi"
    prefixes = [
        "doi",
        "shortdoi",
    ]
    accession_pattern = r"10\.[0-9]{4,9}/\S+"
    shortdoi_pattern = r"10/[a-zA-Z0-9]+"

    def inspect(self, citekey):
        identifier = citekey.accession
        if identifier.startswith("10."):
            # https://www.crossref.org/blog/dois-and-matching-regular-expressions/
            if not self._get_pattern().fullmatch(identifier):
                return (
                    "Identifier does not conform to the DOI regex. "
                    "Double check the DOI."
                )
        elif identifier.startswith("10/"):
            # shortDOI, see http://shortdoi.org
            if not self._get_pattern("shortdoi_pattern").fullmatch(identifier):
                return (
                    "Identifier does not conform to the shortDOI regex. "
                    "Double check the shortDOI."
                )
        else:
            return "DOIs must start with '10.' (or '10/' for shortDOIs)."

    def standardize_prefix_accession(self, accession):
        if accession.startswith("10/"):
            try:
                accession = expand_short_doi(accession)
            except Exception as error:
                # If DOI shortening fails, return the unshortened DOI.
                # DOI metadata lookup will eventually fail somewhere with
                # appropriate error handling, as opposed to here.
                logging.error(
                    f"Error in expand_short_doi for {accession} "
                    f"due to a {error.__class__.__name__}:\n{error}"
                )
                logging.info(error, exc_info=True)
        accession = accession.lower()
        return self.standard_prefix, accession

    def get_csl_item(self, citekey):
        return get_doi_csl_item(citekey.standard_accession)


def expand_short_doi(short_doi: str) -> str:
    """
    Convert a shortDOI to a regular DOI.
    """
    if not short_doi.startswith("10/"):
        raise ValueError(
            f"shortDOIs start with `10/`, but expand_short_doi received: {short_doi}"
        )
    url = f"https://doi.org/api/handles/{short_doi.lower()}"
    params = {"type": "HS_ALIAS"}
    response = requests.get(url, params=params)
    # response documentation at https://www.handle.net/proxy_servlet.html
    results = response.json()
    response_code = results.get("responseCode")  # Handle protocol response code
    if response_code == 100:
        raise ValueError(f"Handle not found. Double check short_doi: {short_doi}")
    if response_code == 200:
        raise ValueError(
            f"HS_ALIAS values not found. Double check short_doi: {short_doi}"
        )
    if response_code != 1:
        raise ValueError(
            f"Error response code of {response_code} returned by {response.url}"
        )
    values = results.get("values", [])
    for value in values:
        if value.get("type") == "HS_ALIAS":
            doi = value["data"]["value"]
            return doi.lower()
    raise RuntimeError(
        f'HS_ALIAS value not found by expand_short_doi("{short_doi}")\n'
        f"The following JSON was retrieved from {response.url}:\n"
        + json.dumps(results, indent=2)
    )


def get_short_doi_url(doi: str) -> Optional[str]:
    """
    Get the shortDOI URL for a DOI.
    """
    quoted_doi = urllib.request.quote(doi)
    url = f"http://shortdoi.org/{quoted_doi}?format=json"
    headers = {"User-Agent": get_manubot_user_agent()}
    try:
        response = requests.get(url, headers=headers).json()
        short_doi = response["ShortDOI"]
        short_url = "https://doi.org/" + short_doi[3:]  # Remove "10/" prefix
        return short_url
    except Exception:
        logging.warning(f"shortDOI lookup failed for {doi}", exc_info=True)
        return None


content_negotiation_url_datacite: str = "https://data.crosscite.org"
content_negotiation_url_default: str = "https://doi.org"


def _get_doi_csl_item_negotiation(doi: str, content_negotiation_url: str):
    """
    Use Content Negotiation to retrieve the CSL Item metadata for a DOI.

    content_negotiation_url: base URL to use for content negotiation.
    Options include:
    1. The DataCite CN service <https://data.crosscite.org>
       Documented at <https://support.datacite.org/docs/datacite-content-resolver>
    2. The doi.org CN service <https://doi.org/>
       Documented at <https://citation.crosscite.org/docs.html>
    """
    url = urllib.parse.urljoin(content_negotiation_url, urllib.request.quote(doi))
    header = {
        "Accept": "application/vnd.citationstyles.csl+json",
        "User-Agent": get_manubot_user_agent(),
    }
    response = requests.get(url, headers=header)
    try:
        return response.json()
    except Exception as error:
        logging.error(
            f"Error fetching metadata for doi:{doi}.\n"
            f"Invalid response from {response.url}:\n{response.text}"
        )
        raise error


def get_doi_csl_item_datacite(doi: str):
    """
    As of 2022-01, the DataCite Content Negotiation restricted
    service to just DataCite DOIs, and began returning 404s for Crossref DOIs.
    https://github.com/crosscite/content-negotiation/issues/104
    """
    return _get_doi_csl_item_negotiation(doi, content_negotiation_url_datacite)


def get_doi_csl_item_default(doi: str):
    """
    doi.org content negotiation redirects to the content negotiation service of
    the Registration Agency, e.g. Crossref or DataCite.
    https://github.com/crosscite/content-negotiation/issues/104
    """
    return _get_doi_csl_item_negotiation(doi, content_negotiation_url_default)


def get_doi_csl_item_zotero(doi: str):
    """
    Generate CSL JSON Data for a DOI using Zotero's translation-server.
    """
    from manubot.cite.zotero import get_csl_item

    return get_csl_item(f"doi:{doi}")


def get_doi_csl_item_url(doi: str):
    """
    Generate CSL JSON Data for a DOI using Zotero's translation-server.
    This function converts the DOI to a URL that presumably resolves to the publisher's site.
    Zotero resolves and scrapes data from the resulting webpage.
    """
    from manubot.cite.url import get_url_csl_item_zotero

    return get_url_csl_item_zotero(f"https://doi.org/{doi}")


def augment_get_doi_csl_item(function: Callable[..., Any]):
    """
    Decorator providing edits to the csl_item returned by a get_doi_csl_item_* function.
    """

    def wrapper(doi: str):
        doi = doi.lower()
        csl_item = function(doi)
        csl_item["DOI"] = doi
        csl_item["URL"] = f"https://doi.org/{doi}"
        short_doi_url = get_short_doi_url(doi)
        if short_doi_url:
            csl_item["URL"] = short_doi_url
        try:
            csl_item.update(get_pubmed_ids_for_doi(doi))
        except Exception:
            logging.warning(
                f"Error calling get_pubmed_ids_for_doi for {doi}", exc_info=True
            )
        return csl_item

    return wrapper


@augment_get_doi_csl_item
def get_doi_csl_item(doi: str):
    """
    Generate CSL JSON Data for an DOI.

    This function uses a list of CSL JSON Item metadata retrievers, specified
    by the module-level variable `doi_retrievers`. The methods are attempted
    in order, with this function returning the metadata from the first
    non-failing method.
    """
    # FIXME: this function is repetitive with other get_*_csl_item functions.
    for retriever in doi_retrievers:
        try:
            return retriever(doi)
        except Exception as error:
            logging.warning(
                f"Error in {retriever.__name__} for {doi} "
                f"due to a {error.__class__.__name__}:\n{error}"
            )
            logging.info(error, exc_info=True)
    raise Exception(f"all get_doi_csl_item methods failed for {doi}")


doi_retrievers = [
    get_doi_csl_item_default,
    get_doi_csl_item_url,
    get_doi_csl_item_zotero,
]
"""
On retreiver ordering

1. get_doi_csl_item_default: try DOI metadata first as per the Registration Agency's Content Negotiation.
2. get_doi_csl_item_url: scrape data from the website where the DOI resolves.
3. get_doi_csl_item_zotero: use Zotero translation-server to get DOI metadata.
   Placed last since it's unlikely to work when get_doi_csl_item_default fails,
   since it also uses content negotiation.

get_doi_csl_item_datacite is not included because as of 2022-01 it only works for DataCite DOIs,
and get_doi_csl_item_default ends up redirecting to this endpoint for DataCite registered DOIs.
"""



import json
import logging
import os
import pathlib

from manubot.cite.citekey import shorten_citekey
from manubot.util import read_serialized_data


def load_bibliography(path: str) -> list:
    """
    Load a bibliography as CSL Items (a CSL JSON Python object).
    For paths that already contain CSL Items (inferred from a .json or .yaml extension),
    parse these files directly (URLs supported).
    Otherwise, delegate conversion to CSL Items to pandoc-citeproc (URLs not supported).
    If loading fails, log an error and return an empty list.
    """
    path_obj = pathlib.Path(path)
    if path_obj.suffix in {".json", ".yaml"}:
        try:
            csl_items = read_serialized_data(path)
        except Exception as error:
            logging.error(f"load_bibliography: error reading {path!r}.\n{error}")
            logging.info("load_bibliography exception info", exc_info=True)
            csl_items = []
    else:
        from manubot.pandoc.bibliography import (
            load_bibliography as load_bibliography_pandoc,
        )

        csl_items = load_bibliography_pandoc(path)
    if not isinstance(csl_items, list):
        logging.error(
            f"process.load_bibliography: csl_items read from {path} are of type {type(csl_items)}. "
            "Setting csl_items to an empty list."
        )
        csl_items = []
    from manubot.cite.csl_item import CSL_Item

    csl_items = [CSL_Item(csl_item) for csl_item in csl_items]
    return csl_items


def load_manual_references(paths=None, extra_csl_items=None) -> dict:
    """
    Read manual references from bibliography text files specified by a list of paths.
    Returns a standard_citation to CSL Item dictionary.
    `extra_csl_items` specifies CSL Items stored as a Python object,
    to be used in addition to the CSL Items stored as text in the files specified by `paths`.
    Set `paths=[]` to only use extra_csl_items.
    When multiple references have the same standard_id,
    precedence is given to reference defined last.
    References in `extra_csl_items` take precedence over those from `paths`.
    """
    if paths is None:
        paths = []
    if extra_csl_items is None:
        extra_csl_items = []

    from manubot.cite.csl_item import CSL_Item

    csl_items = []
    paths = list(dict.fromkeys(paths))  # remove duplicates
    for path in paths:
        path = os.fspath(path)
        path_obj = pathlib.Path(path)
        bibliography = load_bibliography(path)
        for csl_item in bibliography:
            csl_item.note_append_text(
                "Loaded from an external bibliography file by Manubot."
            )
            csl_item.note_append_dict({"source_bibliography": path_obj.name})
            csl_items.append(csl_item)
    csl_items.extend(map(CSL_Item, extra_csl_items))
    manual_refs = {}
    for csl_item in csl_items:
        try:
            csl_item.standardize_id()
        except Exception:
            csl_item_str = json.dumps(csl_item, indent=2)
            logging.info(
                f"Skipping csl_item where setting standard_id failed:\n{csl_item_str}",
                exc_info=True,
            )
            continue
        standard_id = csl_item["id"]
        csl_item.set_id(shorten_citekey(standard_id))
        csl_item.clean()
        manual_refs[standard_id] = csl_item
    return manual_refs



import logging
import re
import xml.etree.ElementTree

import requests

from manubot.util import get_manubot_user_agent

from .csl_item import CSL_Item
from .handlers import Handler


class Handler_arXiv(Handler):

    standard_prefix = "arxiv"

    prefixes = [
        "arxiv",
    ]
    accession_pattern = re.compile(
        r"(?P<versionless_id>[0-9]{4}\.[0-9]{4,5}|[a-z\-]+(\.[A-Z]{2})?/[0-9]{7})(?P<version>v[0-9]+)?"
    )

    def inspect(self, citekey):
        # https://arxiv.org/help/arxiv_identifier
        if not self._get_pattern().fullmatch(citekey.accession):
            return "arXiv identifiers must conform to syntax described at https://arxiv.org/help/arxiv_identifier."

    def get_csl_item(self, citekey):
        return get_arxiv_csl_item(citekey.standard_accession)


class CSL_Item_arXiv(CSL_Item):
    def _set_invariant_fields(self):
        # Set journal/publisher to arXiv
        self["container-title"] = "arXiv"
        self["publisher"] = "arXiv"
        # Set CSL type to report for preprint
        self["type"] = "report"
        return self

    def log_journal_doi(self, arxiv_id, journal_ref=None):
        if "DOI" not in self:
            return
        msg = f"arXiv article {arxiv_id} published at https://doi.org/{self['DOI']}"
        if journal_ref:
            msg += f" — {journal_ref}"
        logging.info(msg)

    def set_identifier_fields(self, arxiv_id):
        self.set_id(f"arxiv:{arxiv_id}")
        self["URL"] = f"https://arxiv.org/abs/{arxiv_id}"
        self["number"] = arxiv_id
        _, version = split_arxiv_id_version(arxiv_id)
        if version:
            self["version"] = version


def split_arxiv_id_version(arxiv_id: str):
    """
    Return (versionless_id, version) tuple.
    Version refers to the verion suffix like 'v2' or None.
    """
    match = re.match(Handler_arXiv.accession_pattern, arxiv_id)
    return match.group("versionless_id"), match.group("version")


def get_arxiv_csl_item(arxiv_id: str):
    """
    Return csl_item item for an arXiv identifier.
    Chooses which arXiv API to use based on whether arxiv_id
    is versioned, since only one endpoint supports versioning.
    """
    _, version = split_arxiv_id_version(arxiv_id)
    if version:
        return get_arxiv_csl_item_export_api(arxiv_id)
    return get_arxiv_csl_item_oai(arxiv_id)


def query_arxiv_api(url, params):
    headers = {"User-Agent": get_manubot_user_agent()}
    response = requests.get(url, params, headers=headers)
    response.raise_for_status()
    xml_tree = xml.etree.ElementTree.fromstring(response.text)
    return xml_tree


def get_arxiv_csl_item_export_api(arxiv_id):
    """
    Return csl_item item for an arXiv record.

    arxiv_id can be versioned, like `1512.00567v2`, or versionless, like
    `1512.00567`. If versionless, the arXiv API will return metadata for the
    latest version. Legacy IDs, such as `cond-mat/0703470v2`, are also
    supported.

    If arXiv has an associated DOI for the record, a warning is logged to
    alert the user that an alternative version of record exists.

    References:
    - https://arxiv.org/help/api/index
    - http://citeproc-js.readthedocs.io/en/latest/csl-json/markup.html
    - https://github.com/citation-style-language/schema/blob/master/csl-data.json
    """
    xml_tree = query_arxiv_api(
        url="https://export.arxiv.org/api/query",
        params={"id_list": arxiv_id, "max_results": 1},
    )

    # XML namespace prefixes
    prefix = "{http://www.w3.org/2005/Atom}"
    alt_prefix = "{http://arxiv.org/schemas/atom}"

    # Parse XML
    (entry,) = xml_tree.findall(prefix + "entry")

    # Create dictionary for CSL Item
    csl_item = CSL_Item_arXiv()

    # Extract versioned arXiv ID
    url = entry.findtext(prefix + "id")
    pattern = re.compile(r"arxiv.org/abs/(.+)")
    match = pattern.search(url)
    versioned_id = match.group(1)
    csl_item.set_identifier_fields(versioned_id)

    # Extrat CSL title field
    csl_item["title"] = entry.findtext(prefix + "title")

    # Extract CSL date field
    published = entry.findtext(prefix + "published")
    csl_item.set_date(published, variable="issued")

    # Extract authors
    authors = []
    for elem in entry.findall(prefix + "author"):
        name = elem.findtext(prefix + "name")
        author = {"literal": name}
        authors.append(author)
    csl_item["author"] = authors

    csl_item._set_invariant_fields()

    # Extract abstract
    abstract = entry.findtext(prefix + "summary").strip()
    if abstract:
        # remove newlines that were added to wrap abstract
        abstract = remove_newlines(abstract)
        csl_item["abstract"] = abstract

    # Check if the article has been published with a DOI
    doi = entry.findtext(f"{alt_prefix}doi")
    if doi:
        csl_item["DOI"] = doi
        journal_ref = entry.findtext(alt_prefix + "journal_ref")
        csl_item.log_journal_doi(arxiv_id, journal_ref)
    return csl_item


def get_arxiv_csl_item_oai(arxiv_id):
    """
    Generate a CSL Item for an unversioned arXiv identifier
    using arXiv's OAI_PMH v2.0 API <https://arxiv.org/help/oa>.
    This endpoint does not support versioned `arxiv_id`.
    """
    # XML namespace prefixes
    ns_oai = "{http://www.openarchives.org/OAI/2.0/}"
    ns_arxiv = "{http://arxiv.org/OAI/arXiv/}"

    xml_tree = query_arxiv_api(
        url="https://export.arxiv.org/oai2",
        params={
            "verb": "GetRecord",
            "metadataPrefix": "arXiv",
            "identifier": f"oai:arXiv.org:{arxiv_id}",
        },
    )

    # Create dictionary for CSL Item
    csl_item = CSL_Item_arXiv()
    # Extract parent XML elements
    (header_elem,) = xml_tree.findall(
        f"{ns_oai}GetRecord/{ns_oai}record/{ns_oai}header"
    )
    (metadata_elem,) = xml_tree.findall(
        f"{ns_oai}GetRecord/{ns_oai}record/{ns_oai}metadata"
    )
    (arxiv_elem,) = metadata_elem.findall(f"{ns_arxiv}arXiv")
    # Set identifier fields
    response_arxiv_id = arxiv_elem.findtext(f"{ns_arxiv}id")
    if arxiv_id != response_arxiv_id:
        logging.warning(
            "arXiv oai2 query returned a different arxiv_id:"
            f" {arxiv_id} became {response_arxiv_id}"
        )
    csl_item.set_identifier_fields(response_arxiv_id)
    # Set title and date
    title = arxiv_elem.findtext(f"{ns_arxiv}title")
    if title:
        csl_item["title"] = " ".join(title.split())
    datestamp = header_elem.findtext(f"{ns_oai}datestamp")
    csl_item.set_date(datestamp, "issued")

    # Extract authors
    author_elems = arxiv_elem.findall(f"{ns_arxiv}authors/{ns_arxiv}author")
    authors = []
    for author_elem in author_elems:
        author = {}
        given = author_elem.findtext(f"{ns_arxiv}forenames")
        family = author_elem.findtext(f"{ns_arxiv}keyname")
        if given:
            author["given"] = given
        if family:
            author["family"] = family
        authors.append(author)
    csl_item["author"] = authors

    csl_item._set_invariant_fields()

    abstract = arxiv_elem.findtext(f"{ns_arxiv}abstract")
    if abstract:
        csl_item["abstract"] = remove_newlines(abstract)

    license = arxiv_elem.findtext(f"{ns_arxiv}license")
    if license:
        csl_item.note_append_dict({"license": license})

    doi = arxiv_elem.findtext(f"{ns_arxiv}doi")
    if doi:
        csl_item["DOI"] = doi
        journal_ref = arxiv_elem.findtext(f"{ns_arxiv}journal-ref")
        csl_item.log_journal_doi(arxiv_id, journal_ref)
    return csl_item


def remove_newlines(text):
    return re.sub(pattern=r"\n(?!\s)", repl=" ", string=text)


def get_arxiv_csl_item_zotero(arxiv_id):
    """
    Generate CSL JSON Data for an arXiv ID using Zotero's translation-server.
    """
    from manubot.cite.zotero import get_csl_item

    return get_csl_item(f"arxiv:{arxiv_id}")



"""
Compact Uniform Resource Identifiers

Manubot keeps a local versions of the Bioregistry.
Repository developers can run the following commands to update the Manubot version.

```shell
# regenerate manubot/cite/curie/bioregistry.json
python manubot/cite/curie/__init__.py
# if bioregistry.json has changed, the following test will likely fail:
pytest manubot/cite/tests/test_handlers.py::test_prefix_to_handler
# copy captured stdout from failed test_prefix_to_handler to
# manubot.cite.handlers.prefix_to_handler. Pre-commit hook will reformat file.
```

References:

- https://bioregistry.io/
- https://github.com/biopragmatics/bioregistry
- https://en.wikipedia.org/wiki/CURIE
- https://cthoyt.com/2021/10/07/biopragmatics-glossary.html
- https://github.com/manubot/manubot/issues/305
- https://identifiers.org/
- https://github.com/manubot/manubot/issues/218
- https://docs.identifiers.org/articles/api.html
- https://n2t.net/e/compact_ids.html
- https://n2t.net/e/cdl_ebi_prefixes.yaml
- https://en.wikipedia.org/wiki/MIRIAM_Registry
- [Identifiers.org and MIRIAM Registry: community resources to provide persistent identification](https://doi.org/10.1093/nar/gkr1097)
- [On the road to robust data citation](https://doi.org/10.1038/sdata.2018.95)
- [Uniform Resolution of Compact Identifiers for Biomedical Data](https://doi.org/10.1038/sdata.2018.29)
"""
import dataclasses
import functools
import json
import pathlib
import re
import typing

from manubot.cite.citekey import CiteKey
from manubot.cite.handlers import Handler

_keep_bioregistry_fields = {
    "deprecated",
    "example",
    "uri_format",
    "name",
    "pattern",
    "preferred_prefix",
    "synonyms",
}


bioregistry_path = pathlib.Path(__file__).parent.joinpath("bioregistry.json")


valid_prefix_pattern = re.compile(r"^[a-z0-9][a-z0-9._-]+?$")
"""
Ignore Bioregistry prefixes/synonyms that do not adhere to this pattern.
More permissive than the pattern at <https://github.com/biopragmatics/bioregistry/issues/158>,
because there are existing Bioregistry prefixes, even preferred prefixes, that are bad.
For example, starting with a number. We primarily care whether the prefix will work as part
of a pandoc citation key without requiring escaping.
"""


@dataclasses.dataclass
class Handler_CURIE(Handler):
    def __post_init__(self):
        try:
            self.resource = get_prefix_to_resource()[self.prefix_lower]
        except KeyError as err:
            raise ValueError(f"Unrecognized CURIE prefix {self.prefix_lower}") from err
        self.standard_prefix = (
            self.resource.get("preferred_prefix") or self.resource["prefix"]
        )
        self.prefixes = self.resource["all_prefixes"]
        if "pattern" in self.resource:
            self.accession_pattern = self.resource["pattern"]

    def get_csl_item(self, citekey: CiteKey):
        from ..url import get_url_csl_item

        url = self.get_url(accession=citekey.standard_accession)
        return get_url_csl_item(url)

    def inspect(self, citekey: CiteKey) -> typing.Optional[str]:
        pattern = self._get_pattern("accession_pattern")
        if pattern and not pattern.fullmatch(citekey.accession):
            return f"{citekey.accession} does not match regex {pattern.pattern}"

    def get_url(self, accession: str) -> str:
        if "uri_format" in self.resource:
            return self.resource["uri_format"].replace("$1", accession)
        return f"https://bioregistry.io/{self.standard_prefix}:{accession}"


def get_curie_handlers():
    """Get all possible CURIE handlers"""
    registries = get_bioregistry(compile_patterns=True)
    handlers = [Handler_CURIE(reg["prefix"]) for reg in registries]
    return handlers


def _download_bioregistry() -> None:
    """
    Download the Bioregistry consensus registry adding the following fields for each registry:
    - prefix: the standard lowercase registry prefix
    - all_prefixes: all distinct valid lowercase prefixes including synonyms
    """
    import requests

    url = "https://github.com/biopragmatics/bioregistry/raw/main/exports/registry/registry.json"
    response = requests.get(url)
    response.raise_for_status()
    results = response.json()
    assert isinstance(results, dict)
    registry = []
    for prefix, resource in results.items():
        assert isinstance(resource, dict)
        if not resource.get("uri_format"):
            # discard unresolvable prefixes
            continue
        for field in set(resource) - _keep_bioregistry_fields:
            del resource[field]
        resource["prefix"] = prefix
        all_prefixes = {
            prefix,
            *(x.lower() for x in resource.pop("synonyms", [])),
        }
        # remove invalid prefixes as per https://github.com/manubot/manubot/pull/306#discussion_r744125504
        resource["all_prefixes"] = sorted(
            filter(valid_prefix_pattern.fullmatch, all_prefixes)
        )
        registry.append(resource)
    json_text = json.dumps(registry, indent=2, ensure_ascii=False)
    bioregistry_path.write_text(json_text + "\n", encoding="utf-8")


def get_bioregistry(compile_patterns=False) -> dict:
    with bioregistry_path.open(encoding="utf-8-sig") as read_file:
        registry = json.load(read_file)
    assert isinstance(registry, list)
    if compile_patterns:
        for resource in registry:
            if "pattern" in resource:
                resource["compiled_pattern"] = re.compile(resource["pattern"])
    return registry


@functools.lru_cache()
def get_prefix_to_resource() -> typing.Dict[str, typing.Dict]:
    prefix_to_resource = {}
    for resource in get_bioregistry():
        for prefix in resource["all_prefixes"]:
            prefix_to_resource[prefix] = resource
    return prefix_to_resource


def standardize_curie(curie: str) -> str:
    """
    Return CURIE with Bioregistry preferred prefix capitalization.
    `curie` should be in `prefix:accession` format.
    If `curie` is malformed or uses an unrecognized prefix, raise ValueError.
    """
    if not isinstance(curie, str):
        raise TypeError(
            f"curie parameter should be string. Received {curie.__class__.__name__} instead for {curie}"
        )
    try:
        prefix, accession = curie.split(":", 1)
    except ValueError as err:
        raise ValueError(
            f"curie must be splittable by `:` and formatted like `prefix:accession`. Received {curie}"
        ) from err
    handler = Handler_CURIE(prefix.lower())
    return f"{handler.standard_prefix}:{accession}"


def curie_to_url(curie: str) -> str:
    """
    `curie` should be in `prefix:accession` format
    """
    curie = standardize_curie(curie)
    prefix, accession = curie.split(":", 1)
    handler = Handler_CURIE(prefix.lower())
    return handler.get_url(accession)


if __name__ == "__main__":
    _download_bioregistry()
    bioregistry = get_bioregistry()



"""
Utilities for accessing <https://unpaywall.org/> data to provide access
information for DOIs. Also supports identifier sources not directly
supported by Unpaywall, such as arXiv IDs.
"""
import abc

import requests

"""
Unpaywall license choices used by Location.has_open_license.
Defaults to licenses that conform to <https://opendefinition.org/>.
"""
open_licenses = {"cc0", "cc-by", "cc-by-sa", "pd"}


class Unpaywall:
    """
    A class to handle open access locations in the Unpaywall data format.
    Create new Unpaywall objects using the `from_csl_item` and `from_citekey` methods,
    or by using __init__ of a subclass like `Unpaywall_DOI`.
    """

    csl_item = None

    @abc.abstractmethod
    def set_oa_locations(self):
        """
        Set `self.oa_locations`, which is a list of `Unpaywall_Location` objects.
        """
        self.oa_locations = []

    @property
    def best_openly_licensed_pdf(self) -> "Unpaywall_Location":
        for location in self.oa_locations:
            if location.has_openly_licensed_pdf:
                return location

    @property
    def best_pdf(self) -> "Unpaywall_Location":
        for location in self.oa_locations:
            if location.has_pdf:
                return location

    @staticmethod
    def from_citekey(citekey, csl_item=None):
        """
        Create an Unpaywall object for `citekey`.
        `csl_item` is an optional field that can avoid an
        external web request to generate to a new CSL Item.
        """
        from .citekey import CiteKey

        if isinstance(citekey, str):
            citekey = CiteKey(citekey)
        if not isinstance(citekey, CiteKey):
            raise ValueError("citekey must be a str or CiteKey")
        if citekey.standard_prefix in source_to_unpaywaller:
            unpaywaller = source_to_unpaywaller[citekey.standard_prefix]
            unpaywall = unpaywaller(citekey.standard_accession, set_oa_locations=False)
        else:
            raise ValueError(
                f"Cannot Unpaywall {citekey.input_id}. "
                f"Supported citations sources are {', '.join(source_to_unpaywaller)}. "
                "Received {citekey.standard_prefix!r}."
            )
        unpaywall.csl_item = csl_item
        unpaywall.set_oa_locations()
        return unpaywall

    @classmethod
    def from_csl_item(cls, csl_item):
        """
        Create an Unpaywall object for `csl_item`.
        """
        from .csl_item import CSL_Item

        csl_item = CSL_Item(csl_item)
        doi = csl_item.get("DOI")
        if doi:
            return cls.from_citekey(f"doi:{doi}", csl_item=csl_item)
        csl_item.infer_id()
        return cls.from_citekey(csl_item["id"], csl_item=csl_item)


class Unpaywall_DOI(Unpaywall):
    """
    From https://unpaywall.org/data-format:

    > The DOI object is more or less a row in our main database...
    it's everything we know about a given DOI-assigned resource,
    including metadata about the resource itself,
    and information about its OA status.
    It includes a list of zero or more OA Location Objects,
    as well as a `best_oa_location` property that's probably the OA Location you'll want to use.
    """

    def __init__(self, doi, set_oa_locations=True):
        self.doi = doi.lower()
        if set_oa_locations:
            self.set_oa_locations()

    def set_oa_locations(self):
        from manubot.util import contact_email

        url = f"https://api.unpaywall.org/v2/{self.doi}"
        params = {"email": contact_email}
        response = requests.get(url, params=params)
        response.raise_for_status()
        self.results = response.json()
        self.oa_locations = [
            Unpaywall_Location(location)
            for location in self.results.get("oa_locations", [])
        ]


class Unpaywall_arXiv(Unpaywall):
    def __init__(self, arxiv_id, set_oa_locations=True, use_doi=True):
        from .arxiv import split_arxiv_id_version

        self.arxiv_id = arxiv_id
        self.arxiv_id_latest, self.arxiv_id_version = split_arxiv_id_version(arxiv_id)
        self.use_doi = use_doi
        if set_oa_locations:
            self.set_oa_locations()

    def set_oa_locations(self):
        from .arxiv import get_arxiv_csl_item

        if not self.csl_item:
            self.csl_item = get_arxiv_csl_item(self.arxiv_id)
        doi = self.csl_item.get("DOI")
        if self.use_doi and doi:
            unpaywall_doi = Unpaywall_DOI(doi)
            self.doi = unpaywall_doi.doi
            self.oa_locations = unpaywall_doi.oa_locations
            return
        location = self.location_from_arvix_id()
        self.oa_locations = [location]

    def location_from_arvix_id(self):
        import datetime

        url_for_pdf = f"https://arxiv.org/pdf/{self.arxiv_id}.pdf"
        location = Unpaywall_Location(
            {
                "endpoint_id": None,
                "evidence": "oa repository",
                "host_type": "repository",
                "is_best": True,
                "license": self.get_license(),
                "pmh_id": f"oai:arXiv.org:{self.arxiv_id_latest}",
                "repository_institution": "Cornell University - arXiv",
                "updated": datetime.datetime.now().isoformat(),
                "url": url_for_pdf,
                "url_for_landing_page": f"https://arxiv.org/abs/{self.arxiv_id}",
                "url_for_pdf": url_for_pdf,
                "version": "submittedVersion",
            }
        )
        return location

    def get_license(self):
        """
        Return license using choices from the Unpaywall data format.
        Looks for license metadata in the CSL Item.
        """
        license = self.csl_item.note_dict.get("license")
        if not license:
            return
        # Example licenses from https://arxiv.org/help/license
        # http://creativecommons.org/publicdomain/zero/1.0/
        # http://creativecommons.org/licenses/by/4.0/
        # http://creativecommons.org/licenses/by-sa/4.0/
        # http://creativecommons.org/licenses/by-nc-sa/4.0/
        # http://arxiv.org/licenses/nonexclusive-distrib/1.0/license.html
        from urllib.parse import urlparse

        parsed_url = urlparse(license)
        if not parsed_url.scheme.startswith("http"):
            return
        if parsed_url.hostname.endswith("creativecommons.org"):
            try:
                abbrev = parsed_url.path.split("/")[2]
            except IndexError:
                return
            if abbrev == "zero":
                return "cc0"
            return f"cc-{abbrev}"


source_to_unpaywaller = {
    "doi": Unpaywall_DOI,
    "arxiv": Unpaywall_arXiv,
}


class Unpaywall_Location(dict):
    """
    From https://unpaywall.org/data-format

    > The OA Location object describes particular place where we found a given OA article.
    The same article is often available from multiple locations,
    and there may be differences in format, version, and license depending on the location;
    the OA Location object describes these key attributes.
    An OA Location Object is always a Child of a DOI Object.

    Example oa_locations from the Unpaywall API are:

    ```json
    {
        "endpoint_id": null,
        "evidence": "open (via page says license)",
        "host_type": "publisher",
        "is_best": true,
        "license": "cc-by",
        "pmh_id": null,
        "repository_institution": null,
        "updated": "2020-01-19T08:55:45.548214",
        "url": "https://journals.plos.org/ploscompbiol/article/file?id=10.1371/journal.pcbi.1007250&type=printable",
        "url_for_landing_page": "https://doi.org/10.1371/journal.pcbi.1007250",
        "url_for_pdf": "https://journals.plos.org/ploscompbiol/article/file?id=10.1371/journal.pcbi.1007250&type=printable",
        "version": "publishedVersion"
    },
    {
        "endpoint_id": "ca8f8d56758a80a4f86",
        "evidence": "oa repository (via OAI-PMH doi match)",
        "host_type": "repository",
        "is_best": true,
        "license": null,
        "pmh_id": "oai:arXiv.org:1806.05726",
        "repository_institution": "Cornell University - arXiv",
        "updated": "2019-11-01T00:28:16.784912",
        "url": "http://arxiv.org/pdf/1806.05726",
        "url_for_landing_page": "http://arxiv.org/abs/1806.05726",
        "url_for_pdf": "http://arxiv.org/pdf/1806.05726",
        "version": "submittedVersion"
    }
    ```
    """

    @property
    def has_pdf(self):
        return bool(self.get("url_for_pdf"))

    @property
    def has_open_license(self):
        license = self.get("license")
        return license in open_licenses

    @property
    def has_creative_commons_license(self):
        license = self.get("license")
        if not license:
            return False
        return license == "cc0" or license.startswith("cc-")

    @property
    def has_openly_licensed_pdf(self):
        return self.has_pdf and self.has_open_license



import datetime
import json
import logging
import pathlib


def get_text(directory):
    """
    Return a concatenated string of section texts from the specified directory.
    Text files should be UTF-8 encoded.
    """
    section_dir = pathlib.Path(directory)
    paths = sorted(section_dir.glob("[0-9]*.md"))
    name_to_text = {}
    for path in paths:
        name_to_text[path.stem] = path.read_text(encoding="utf-8-sig")
    logging.info("Manuscript content parts:\n" + "\n".join(name_to_text))
    return "\n\n".join(name_to_text.values()) + "\n"


def get_manuscript_stats(text):
    """
    Compute manuscript statistics.
    """
    stats = {}
    stats["word_count"] = len(text.split())
    logging.info(f"Generated manscript stats:\n{json.dumps(stats, indent=2)}")
    return stats


def datetime_now():
    """
    Return the current datetime, with timezone awareness
    https://stackoverflow.com/a/39079819/4651668
    """
    tzinfo = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
    return datetime.datetime.now(tzinfo)



from typing import Any, Dict

from .handlers import Handler


class Handler_Wikidata(Handler):
    prefixes = ["wikidata"]
    standard_prefix = "wikidata"
    accession_pattern = r"Q[0-9]+"

    def inspect(self, citekey):
        """
        https://www.wikidata.org/wiki/Wikidata:Identifiers
        """
        accession = citekey.accession
        if not accession.startswith("Q"):
            return "Wikidata item IDs must start with 'Q'."
        elif not self._get_pattern().fullmatch(accession):
            return (
                "Accession does not conform to the Wikidata regex. "
                "Double check the entity ID."
            )

    def get_csl_item(self, citekey):
        return get_wikidata_csl_item(citekey.standard_accession)


def get_wikidata_csl_item(identifier: str) -> Dict[str, Any]:
    """
    Get a CSL JSON item with the citation metadata for a Wikidata item.
    identifier should be a Wikidata item ID corresponding to a citeable
    work, such as Q50051684.
    """
    url = f"https://www.wikidata.org/wiki/{identifier}"
    from manubot.cite.url import get_url_csl_item_zotero

    csl_item = get_url_csl_item_zotero(url)
    if "DOI" in csl_item:
        csl_item["DOI"] = csl_item["DOI"].lower()
    if "URL" not in csl_item:
        csl_item["URL"] = url
    return csl_item



"""
Tools for manuscript metadata processing including thumbnail detection and processing.
"""
import functools
import logging
import pathlib
import subprocess
from typing import Optional
from urllib.parse import urljoin


def get_header_includes(variables: dict) -> str:
    """
    Render `header-includes-template.html` using information from `variables`.
    """
    from .util import template_with_jinja2

    path = pathlib.Path(__file__).parent.joinpath("header-includes-template.html")
    try:
        template = path.read_text(encoding="utf-8-sig")
        return template_with_jinja2(template, variables)
    except Exception:
        logging.exception("Error generating header-includes.")
        return ""


def get_thumbnail_url(thumbnail=None):
    """
    Starting with a user-specified `thumbnail` as either a path, URL, or None,
    return an absolute URL pointing to the thumbnail image. If the provided `thumbnail`
    is a URL, return this URL unmodified. If `thumbnail` is None, search for `thumbnail.png`
    within the git repository from which this function is executed. If `thumbnail`
    is a local path, the path should be relative to root directory of the git repository
    it is located in. If a local path is provided or detected,
    it is converted to a GitHub raw URL.
    """
    from manubot.util import is_http_url

    if not thumbnail:
        message = "get_thumbnail_url: thumbnail location not explicitly provided. "
        thumbnail = _find_thumbnail_path()
        message += (
            f"Thumbnail detected at {thumbnail!r}"
            if thumbnail
            else "No local thumbnail detected"
        )
        logging.debug(message)
    elif is_http_url(thumbnail):
        logging.debug("provided thumbnail is a URL. Pass it through.")
        return thumbnail
    return _thumbnail_path_to_url(thumbnail)


def _find_thumbnail_path():
    """
    If this this function is executed with a working directory that is inside a git repository,
    return the path to a `thumbnail.png` file located anywhere in that repository. Otherwise,
    return `None`.
    """
    directory = git_repository_root()
    if not directory:
        return None
    paths = directory.glob("**/thumbnail.png")
    paths = [path.relative_to(directory) for path in paths]
    paths = sorted(paths, key=lambda x: (len(x.parents), x))
    if not paths:
        return None
    return paths[0].as_posix()


def _thumbnail_path_to_url(path):
    """
    Convert a local thumbnail path (string) to an absolute URL using the GitHub
    repository location detected using `get_continuous_integration_parameters`.
    """
    if not path:
        return None
    from .ci import get_continuous_integration_parameters

    info = get_continuous_integration_parameters()
    try:
        url = f"https://github.com/{info['repo_slug']}/raw/{info['triggering_commit']}/{path}"
    except (TypeError, KeyError):
        return None
    return url


def get_head_commit() -> Optional[str]:
    from manubot.util import shlex_join

    args = ["git", "rev-parse", "HEAD"]
    try:
        return subprocess.check_output(args, stderr=subprocess.PIPE, text=True).strip()
    except subprocess.CalledProcessError as error:
        logging.warning(
            f"get_head_commit: {shlex_join(error.cmd)!r} returned exit code {error.returncode} "
            f"with the following stdout:\n{error.stdout}\n"
            f"And the following stderr:\n{error.stderr}"
        )
        return None


@functools.lru_cache()
def git_repository_root():
    """
    Return the path to repository root directory or `None` if indeterminate.
    """
    for cmd in (
        ["git", "rev-parse", "--show-superproject-working-tree"],
        ["git", "rev-parse", "--show-toplevel"],
    ):
        try:
            path = subprocess.check_output(cmd, text=True).rstrip("\r\n")
            if path:
                return pathlib.Path(path)
        except (subprocess.CalledProcessError, OSError):
            pass
    return None


def get_manuscript_urls(html_url: Optional[str] = None) -> dict:
    """
    Return a dictionary with URLs for a manuscript.
    An example for a manuscript where all URLs get set, inferred from continuous integration environment variables, is:
    ```python
    {
        "html_url": "https://manubot.github.io/rootstock/",
        "pdf_url": "https://manubot.github.io/rootstock/manuscript.pdf",
        "html_url_versioned": "https://manubot.github.io/rootstock/v/7cf9071212ce33116ad09cf2237a370b180a3c35/",
        "pdf_url_versioned": "https://manubot.github.io/rootstock/v/7cf9071212ce33116ad09cf2237a370b180a3c35/manuscript.pdf",
    }
    ```

    Provide `html_url` to set a custom domain.
    If `html_url="https://git.dhimmel.com/bitcoin-whitepaper/"`,
    the return dictionary will be like:
    ```python
    {
        "html_url": "https://git.dhimmel.com/bitcoin-whitepaper/",
        "pdf_url": "https://git.dhimmel.com/bitcoin-whitepaper/manuscript.pdf",
        "html_url_versioned": "https://git.dhimmel.com/bitcoin-whitepaper/v/cb1f2c12eec8b56db9ef5f641ec805e2d449d319/",
        "pdf_url_versioned": "https://git.dhimmel.com/bitcoin-whitepaper/v/cb1f2c12eec8b56db9ef5f641ec805e2d449d319/manuscript.pdf",
    }
    ```
    Note the trailing `/` in `html_url`, which is required for proper functioning.
    """
    import requests

    from .ci import get_continuous_integration_parameters

    urls = {}
    ci_params = get_continuous_integration_parameters()
    if html_url is None:
        if not ci_params:
            return urls
        html_url = "https://{repo_owner}.github.io/{repo_name}/".format(**ci_params)
    urls["html_url"] = html_url
    urls["pdf_url"] = urljoin(html_url, "manuscript.pdf")
    if not ci_params:
        return urls
    urls["html_url_versioned"] = urljoin(html_url, "v/{commit}/".format(**ci_params))
    urls["pdf_url_versioned"] = urljoin(urls["html_url_versioned"], "manuscript.pdf")
    response = requests.head(html_url, allow_redirects=True)
    if not response.ok:
        logging.warning(
            "html_url is not web accessible. "
            f"{html_url} returned status code {response.status_code}. "
            "Ignore this warning if the manuscript has not yet been deployed for the first time. "
        )
    if response.history:
        logging.info(
            "html_url includes redirects. In order of oldest to most recent:\n"
            + "\n".join(x.url for x in response.history + [response])
        )
    return urls


def get_software_versions(rootstock: bool = True) -> dict:
    """
    Return a dictionary of software versions for softwares components:

    - manubot_version: the semantic version number of the manubot python package.
    - rootstock_commit: the version of the rootstock repository, as a commit hash,
      included in the manuscript repository.

    Values whose detection fails are set to None.

    The `rootstock` parameter controls whether to fetch the rootstock commit id.
    The rootstock git remote will be added to the local git repository if true.
    """
    from manubot import __version__ as manubot_version

    return {
        "manubot_version": manubot_version,
        "rootstock_commit": get_rootstock_commit() if rootstock else None,
    }


def get_rootstock_commit() -> Optional[str]:
    """
    Return the most recent commit in common between the git repository
    this function is run within (usually a Manubot manuscript repository)
    and the `main` branch of the `rootstock` remote.

    WARNING: This function may modify the git repository its executed within:

    - if the repository has not set the `roostock` remote, it is set to
      point to the default Rootstock repository of <https://github.com/manubot/rootstock>.
    - fetches the latest commits in the `main` branch of the `rootstock` remote
    """
    from manubot.util import shlex_join

    # add rootstock remote if remote is not already set
    rootstock_remote = "https://github.com/manubot/rootstock.git"
    args = ["git", "remote", "add", "rootstock", rootstock_remote]
    process = subprocess.run(args, capture_output=True)
    if process.returncode == 0:
        logging.info(
            "get_rootstock_commit added a `rootstock` remote to the git repository."
        )
    # find most recent common ancestor commit
    try:
        args = ["git", "fetch", "rootstock", "main"]
        subprocess.check_output(args, stderr=subprocess.PIPE, text=True)
        args = ["git", "merge-base", "HEAD", "rootstock/main"]
        output = subprocess.check_output(args, stderr=subprocess.PIPE, text=True)
    except subprocess.CalledProcessError as error:
        logging.warning(
            f"get_rootstock_commit: {shlex_join(error.cmd)!r} returned exit code {error.returncode} "
            f"with the following stdout:\n{error.stdout}\n"
            f"And the following stderr:\n{error.stderr}"
        )
        return None
    rootstock_commit = output.strip()
    return rootstock_commit



import logging
import os
import pathlib

import requests
import requests_cache


class RequestsCache:
    def __init__(self, path):
        self.path = os.fspath(path)

    def mkdir(self):
        """make directory containing cache file if it doesn't exist"""
        directory = pathlib.Path(self.path).parent
        directory.mkdir(parents=True, exist_ok=True)

    def install(self):
        """install cache"""
        requests  # require `import requests` in case this is essential for monkey patching by requests_cache.
        requests_cache.install_cache(self.path, include_get_headers=True)
        self.cache = requests_cache.get_cache()
        logging.info(
            f"requests-cache starting with {len(self.cache.responses)} cached responses"
        )

    def clear(self):
        """clear cache"""
        logging.info("Clearing requests-cache")
        requests_cache.clear()

    def close(self):
        """uninstall cache"""
        logging.info(
            f"requests-cache finished with {len(self.cache.responses)} cached responses"
        )
        requests_cache.uninstall_cache()
